from allauth.socialaccount.models import SocialApp
from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand

from fedeproxy import settings
from fedeproxy.common.gitlab import GitLab


class Command(BaseCommand):
    help = "Create an API user"

    def add_arguments(self, parser):
        parser.add_argument("--config", default=f"{settings.CONFIG_DIR}")
        parser.add_argument("domain")
        parser.add_argument("username")
        parser.add_argument("password")

    def setup_gitlab_auth(self, *args, **kwargs):
        provider = "gitlab"
        name = "GitLab"

        existing = SocialApp.objects.filter(provider=provider)
        if existing:
            return False

        gitlab = GitLab(f'https://{kwargs["domain"]}')
        gitlab.login("root", kwargs["password"])
        gitlab.ensure_group_exists("fedeproxy", request_access_enabled=True, visibility="public")
        (client_id, client_secret) = gitlab.create_api_application(kwargs["domain"])

        a = SocialApp(provider=provider, name=name, secret=client_secret, client_id=client_id, key="")
        a.save()

        # Now associate this provider with all site instances.
        sites = [i for i in Site.objects.all()]
        a.sites.add(*sites)

        return True

    def handle(self, *args, **kwargs):
        changed = False
        changed |= self.setup_gitlab_auth(*args, **kwargs)

        if changed:
            self.stdout.write("Changed")
        else:
            self.stdout.write("Up to date")



import os

import pytest

from fedeproxy.architecture.forge.gitea import Gitea

from . import fedeproxy, identity


@pytest.fixture
def forge(tmpdir):
    f = fedeproxy.Fedeproxy(Gitea, f"http://{os.environ.get('MY_IP', '0.0.0.0')}:8781")
    f.base_directory = f"{tmpdir}/fedeproxy"
    f.init()
    own = f.own
    own.authenticate(username="root", password="Wrobyak4")
    own.project_delete()
    email = "testuser2@example.com"
    password = "Wrobyak4"
    own.user_delete(own.owner)
    own.user_create(own.owner, password, email)
    own.authenticate(username=own.owner, password=password)
    own.init()
    yield own
    own.authenticate(username="root", password="Wrobyak4")
    own.project_delete()
    own.user_delete(own.owner)


@pytest.fixture
def identity_sample():
    emails = ["one@example.com", "two@example.com"]
    url = "url"
    owned = True
    token = "TOKEN"
    return {
        "emails": emails,
        "url": url,
        "owned": owned,
        "token": token,
    }


def test_identity_private(identity_sample):
    i = identity.IdentityPrivate()
    assert i.serialize() == {
        "emails": [],
        "url": None,
        "owned": False,
        "token": None,
    }
    assert i.is_owned() is False
    assert i.is_staged() is True
    i.unserialize(identity_sample)
    assert i.serialize() == identity_sample
    assert i.is_owned() is True
    assert i.is_staged() is False


def test_identity_public(identity_sample):
    i_private = identity.IdentityPrivate()
    i_private.unserialize(identity_sample)
    i_public = identity.IdentityPublic()
    i_public.from_private(i_private)
    expected = {
        "emails": [
            "d25354f658256d3988a0a1f07ae2dbfa64c0141b3eaa5650fd268b3fdd903b7d",
            "1f6c1f35fba8e0f461ef40adaec3cbda883f6a5bcfa5fddef2df80af49fc0832",
        ],
        "owned": True,
        "url": "url",
    }
    assert i_public.serialize() == expected
    i_public.unserialize(expected)
    assert i_public.serialize() == expected


def test_identities(forge, identity_sample):
    identities = identity.IdentitiesPrivate(forge)
    identities.init()
    assert identities.load() is False
    assert identities.identities == []
    i = identity.IdentityPrivate()
    i.unserialize(identity_sample)
    identities.identities.append(i)
    identities.save()

    identities = identity.IdentitiesPrivate(forge)
    identities.init()
    assert identities.load() is True
    identity_loaded = identities.identities[0]
    assert identity_loaded.serialize() == identity_sample



import os

import pytest
import requests

#
# forge.Forge
#


def test_authenticate(forge):
    (Forge, url) = (type(forge), forge.url)
    forgeA = Forge(url)
    forgeA.authenticate(username="root", password="Wrobyak4")
    assert forgeA.user_get("root") is not None
    forgeB = Forge(url)
    forgeB.authenticate(token=forgeA.get_token())
    assert forgeB.user_get("root") is not None


#
# forge.Project
#


def test_project_create(forge, make_user, password):
    username = "testuser1"
    make_user(username)
    forge.authenticate(username=username, password=password)

    forge.project_delete(username, "testproject")
    assert forge.project_get(username, "testproject") is None
    p = forge.project_create(username, "testproject")
    assert p.project == "testproject"
    assert p.id == forge.project_create(username, "testproject").id
    assert forge.project_delete(username, "testproject") is True
    assert forge.project_delete(username, "testproject") is False


@pytest.mark.forges(["GitLab"])
def test_project_save(tmpdir, forge, make_user, password, make_project):
    username = "testuser1"
    make_user(username)

    forge.authenticate(username=username, password=password)
    p = make_project(username, "testproject")
    p.issues.create("ISSUE ONE")
    p.issues.create("ISSUE TWO")
    p.save(tmpdir)
    assert p.load(tmpdir) == 2


@pytest.mark.forges(["GitLab"])
def test_project_dvcs(tmpdir, forge, make_user, password, make_project):
    username = "testuser1"
    make_user(username)

    forge.authenticate(username=username, password=password)
    p = make_project(username, "testproject")

    directory = f"{tmpdir}/somewhere"
    d = p.dvcs(directory)
    assert not os.path.exists(d.directory)
    assert d.clone("master") is True
    assert os.path.exists(d.directory)


#
# forge.User
#


def test_user_create(forge):
    forge.authenticate(username="root", password="Wrobyak4")
    username = "testuser1"
    email = "testuser1@example.com"
    password = "Wrobyak4"
    forge.user_delete(username)
    u = forge.user_create(username, password, email)
    assert u.url == forge.user_create(username, password, email).url
    forge.authenticate(username=username, password=password)
    forge.authenticate(username="root", password="Wrobyak4")
    assert forge.user_delete(username) is True
    assert forge.user_get(username) is None
    assert forge.user_delete(username) is False


def test_user_get(forge):
    forge.authenticate(username="root", password="Wrobyak4")

    password = "Wrobyak4"

    username1 = "testuser1"
    email1 = "testuser1@example.com"
    forge.user_delete(username1)
    forge.user_create(username1, password, email1)

    username2 = "testuser2"
    email2 = "testuser2@example.com"
    forge.user_delete(username2)
    forge.user_create(username2, password, email2)

    #
    # As an admin, the email of all users is exposed
    #
    u = forge.user_get(username1)
    assert u.emails == [email1]

    #
    # When authenticated as a user, its email is exposed
    #
    forge.authenticate(username=username1, password=password)
    u = forge.user_get(username1)
    assert u.emails == [email1]

    #
    # When authenticated as an unprivileged user, its email is unknown
    # and is either not returned at all or replaced by a placeholder
    # that is different from the real email
    #
    forge.authenticate(username=username2, password=password)
    u = forge.user_get(username1)
    assert u.emails == [] or u.emails[0] != email1

    forge.authenticate(username="root", password="Wrobyak4")
    assert forge.user_delete(username1) is True
    assert forge.user_delete(username2) is True


def test_user_url(forge):
    forge.authenticate(username="root", password="Wrobyak4")
    username = "testuser1"
    email = "testuser1@example.com"
    password = "Wrobyak4"
    forge.user_delete(username)
    u = forge.user_create(username, password, email)
    assert u.username == username
    r = requests.get(u.url)
    r.raise_for_status()


#
# forge.Milestones & forge.Milestone
#


@pytest.mark.forges(["Gitea"])
def test_milestone_create(forge, make_user, password, make_project):
    username = "testuser1"
    make_user(username)

    forge.authenticate(username=username, password=password)
    p = make_project(username, "testproject")

    title = "THE TITLE"
    i = p.milestones.create(title)
    assert i.id == p.milestones.get(i.id).id
    assert p.milestones.delete(i.id) is True
    assert p.milestones.get(i.id) is None
    assert p.milestones.delete(i.id) is False


#
# forge.Issues & forge.Issue
#


@pytest.mark.forges(["GitLab"])
def test_issue_create(forge, make_user, password, make_project):
    username = "testuser1"
    make_user(username)

    forge.authenticate(username=username, password=password)
    p = make_project(username, "testproject")

    title = "THE TITLE"
    i = p.issues.create(title)
    assert i.id == p.issues.get(i.id).id
    for x in p.issues.list():
        assert x.id == i.id
    assert p.issues.delete(i.id) is True
    assert p.issues.get(i.id) is None
    assert p.issues.delete(i.id) is False


@pytest.mark.forges(["GitLab"])
def test_issue_save(tmpdir, forge, make_user, password, make_project):
    username = "testuser1"
    make_user(username)

    forge.authenticate(username=username, password=password)
    p = make_project(username, "testproject")

    title = "THE TITLE"
    p.issues.create(title)
    save_path = f"{tmpdir}/issue.json"
    p.issues.save(save_path)
    assert p.issues.load(save_path) == 1



import abc


class Persistent(abc.ABC):
    @abc.abstractmethod
    def save(self, pathname):
        ...

    @abc.abstractmethod
    def load(self, pathname):
        ...


class Forge(abc.ABC):
    def __init__(self, url):
        self._url = url

    @property
    def url(self):
        return self._url

    @property
    def s(self):
        return self._s

    @property
    @abc.abstractmethod
    def username(self):
        ...

    @abc.abstractmethod
    def authenticate(self, **kwargs):
        ...

    @abc.abstractmethod
    def get_token(self):
        ...

    @abc.abstractmethod
    def user_factory(self):
        ...

    @abc.abstractmethod
    def user_get(self, username):
        ...

    @abc.abstractmethod
    def user_create(self, username, password, email):
        ...

    @abc.abstractmethod
    def user_delete(self, username):
        ...

    @abc.abstractmethod
    def project_factory(self):
        ...

    @abc.abstractmethod
    def project_get(self, namespace, project):
        ...

    @abc.abstractmethod
    def project_create(self, namespace, project, **data):
        ...

    @abc.abstractmethod
    def project_delete(self, namespace, project):
        ...


class Project(Persistent):
    def __init__(self, forge, project):
        self._forge = forge
        self._project = project
        self._issues = self.issues_factory()(self)
        self._milestones = self.milestones_factory()(self)

    @property
    def forge(self):
        return self._forge

    @property
    def s(self):
        return self.forge.s

    @property
    @abc.abstractmethod
    def id(self):
        ...

    @property
    @abc.abstractmethod
    def namespace(self):
        ...

    @property
    @abc.abstractmethod
    def project(self):
        ...

    @abc.abstractmethod
    def issues_factory(self):
        ...

    @property
    def issues(self):
        return self._issues

    @abc.abstractmethod
    def milestones_factory(self):
        ...

    @property
    def milestones(self):
        return self._milestones

    @property
    @abc.abstractmethod
    def ssh_url_to_repo(self):
        ...

    @property
    @abc.abstractmethod
    def http_url_to_repo(self):
        ...

    @abc.abstractmethod
    def dvcs_factory(self):
        ...

    @abc.abstractmethod
    def dvcs(self, directory):
        ...


class Milestones(abc.ABC):
    def __init__(self, project):
        self._project = project

    @property
    def project(self):
        return self._project

    @property
    def s(self):
        return self.project.s

    @abc.abstractmethod
    def get(self, id):
        ...

    @abc.abstractmethod
    def create(self, title, **data):
        ...

    @abc.abstractmethod
    def delete(self, id):
        ...

    @abc.abstractmethod
    def list(self):
        ...


class Milestone(abc.ABC):
    def __init__(self, project, milestone):
        self._project = project
        self._milestone = milestone

    @property
    def project(self):
        return self._project

    @property
    def s(self):
        return self.project.s

    @property
    @abc.abstractmethod
    def id(self):
        ...


class Issues(Persistent):
    def __init__(self, project):
        self._project = project

    @property
    def project(self):
        return self._project

    @property
    def s(self):
        return self.project.s

    @abc.abstractmethod
    def get(self, id):
        ...

    @abc.abstractmethod
    def create(self, title, **data):
        ...

    @abc.abstractmethod
    def delete(self, id):
        ...

    @abc.abstractmethod
    def list(self):
        ...


class Issue(abc.ABC):
    def __init__(self, project, issue):
        self._project = project
        self._issue = issue

    @property
    def project(self):
        return self._project

    @property
    def s(self):
        return self.project.s

    @property
    @abc.abstractmethod
    def id(self):
        ...

    @abc.abstractmethod
    def to_json(self):
        ...

    @abc.abstractmethod
    def from_json(self, j):
        ...


class User(abc.ABC):
    def __init__(self, forge, user):
        self._forge = forge
        self._user = user

    @property
    def forge(self):
        return self._forge

    @property
    def s(self):
        return self.forge.s

    @property
    @abc.abstractmethod
    def url(self):
        ...

    @property
    @abc.abstractmethod
    def username(self):
        ...

    @property
    @abc.abstractmethod
    def emails(self):
        ...



import copy
import logging
import time

import gitlab
import requests

from fedeproxy.common.retry import retry

from . import forge

logger = logging.getLogger(__name__)


class GitLab(forge.Forge):
    def __init__(self, url):
        super().__init__(url)

    def authenticate(self, **kwargs):
        if "token" in kwargs:
            self.set_token(kwargs["token"])
        else:
            self.login(kwargs["username"], kwargs["password"])
        self._s = gitlab.Gitlab(url=self.url, oauth_token=self.token)
        self.s.auth()

    @property
    def username(self):
        return self.s.user.username

    @property
    def is_admin(self):
        return self.s.user.is_admin

    def login(self, username, password):
        r = requests.post(
            self.url + "/oauth/token",
            json={
                "username": username,
                "password": password,
                "grant_type": "password",
            },
        )
        logger.debug(r.text)
        r.raise_for_status()
        self.set_token(r.json()["access_token"])

    def get_token(self):
        return self.token

    def set_token(self, token):
        self.token = token

    def get_namespace_id(self, name):
        r = self.s.namespaces.list(search=name)
        return r[0].id

    def user_delete(self, username):
        user = self.user_get(username)
        if user is None:
            return False
        while True:
            try:
                self.s.users.delete(user.id)
            except gitlab.exceptions.GitlabDeleteError as e:
                if e.response_code == 404:
                    break
                raise
            time.sleep(0.1)
        return True

    def user_factory(self):
        return GitLabUser

    def user_get(self, username):
        if username == self.username:
            return self.user_factory()(self, self.s.user)

        for found in self.s.users.list(username=username):
            #
            # There may be more than one match because the search is case insensitive
            #
            if found.username == username:
                return self.user_factory()(self, found)
        return None

    def user_create(self, username, password, email):
        info = self.user_get(username)
        if info is None:
            self.s.users.create(
                {
                    "name": username,
                    "username": username,
                    "email": email,
                    "password": password,
                    "skip_confirmation": True,
                }
            )
            info = self.user_get(username)
        return info

    def project_delete(self, namespace, project):
        p = self.project_get(namespace, project)
        if p is None:
            return False
        self.s.projects.delete(f"{namespace}/{project}")
        while self.project_get(namespace, project) is not None:
            time.sleep(1)
        return True

    def project_factory(self):
        return GitLabProject

    def project_get(self, namespace, project):
        try:
            p = self.s.projects.get(f"{namespace}/{project}")
            return self.project_factory()(self, p)
        except gitlab.exceptions.GitlabGetError as e:
            if e.response_code == 404:
                return None
            raise

    class DeletionInProgress(Exception):
        pass

    @retry(DeletionInProgress, tries=5)
    def _project_create(self, namespace, project, **data):
        namespace_id = self.get_namespace_id(namespace)
        data.update(
            {
                "name": project,
                "namespace_id": int(namespace_id),
                "visibility": "public",
            }
        )
        try:
            self.s.projects.create(data)
        except gitlab.exceptions.GitlabCreateError as e:
            if e.response_code == 400 and (
                "still being deleted" in e.response_body.decode()
                or "has already been taken" in e.response_body.decode()
            ):
                raise GitLab.DeletionInProgress()
            raise
        return self.project_get(namespace, project)

    def project_create(self, namespace, project, **data):
        info = self.project_get(namespace, project)
        if info is None:
            return self._project_create(namespace, project, **data)
        else:
            return info


class GitLabProject(forge.Project):
    @property
    def id(self):
        return self._project.id

    @property
    def namespace(self):
        return self._project.namespace.path

    @property
    def project(self):
        return self._project.path

    def milestones_factory(self):
        return GitLabMilestones

    def issues_factory(self):
        return GitLabIssues

    @property
    def ssh_url_to_repo(self):
        return self._project.ssh_url_to_repo

    @property
    def http_url_to_repo(self):
        return self._project.http_url_to_repo


class GitLabMilestones(forge.Milestones):
    def get(self, id):
        pass

    def delete(self, id):
        pass

    def create(self, title, **data):
        pass

    def list(self):
        pass


class GitLabMilestone(forge.Milestone):
    @property
    def id(self):
        pass


class GitLabIssues(forge.Issues):
    def get(self, id):
        try:
            i = self.project._project.issues.get(id)
            return GitLabIssue(self.project, i)
        except gitlab.exceptions.GitlabGetError as e:
            if e.response_code == 404:
                return None
            raise

    def delete(self, id):
        info = self.get(id)
        if info is None:
            return False
        self.project._project.issues.delete(id)
        return True

    def create(self, title, **data):
        data = copy.copy(data)
        data.update({"title": title})
        i = self.project._project.issues.create(data)
        return self.get(i.iid)

    def list(self):
        for i in self.project._project.issues.list(scope="all"):
            yield GitLabIssue(self.project, i)


class GitLabIssue(forge.Issue):
    @property
    def id(self):
        return self._issue.iid

    def to_json(self):
        i = self._issue
        j = {
            "url": i.web_url,
            "title": i.title,
            "description": i.description or "",
            "repository": self.project.http_url_to_repo,
            "user": i.author["web_url"],
            "created_at": i.created_at,
            "state": i.state == "closed" and "closed" or "open",
            "comments": [],
            "id": str(i.iid),
        }
        if i.closed_at is not None:
            j["closed_at"] = i.closed_at
        return j

    def from_json(self, j):
        self._issue.title = j["title"]
        self._issue.description = j["description"]
        self._issue.state = j["state"]
        self._issue.save()


class GitLabUser(forge.User):
    @property
    def url(self):
        return self._user.web_url

    @property
    def username(self):
        return self._user.username

    @property
    def emails(self):
        if hasattr(self._user, "email"):
            return [self._user.email]
        elif hasattr(self._user, "public_email"):
            return [self._user.public_email]
        else:
            return []

    @property
    def id(self):
        return self._user.id



import argparse

from cliff.command import Command


class Manage(Command):
    "Manage"

    def get_parser(self, prog_name):
        parser = super(Manage, self).get_parser(prog_name)
        parser.add_argument("args", nargs=argparse.REMAINDER)
        return parser

    def take_action(self, parsed_args):
        try:
            from django.core.management import execute_from_command_line
        except ImportError as exc:
            raise ImportError(
                "Couldn't import Django. Are you sure it's installed and "
                "available on your PYTHONPATH environment variable? Did you "
                "forget to activate a virtual environment?"
            ) from exc
        execute_from_command_line(["fedeproxy", *parsed_args.args])



import abc


class DVCS(abc.ABC):
    @abc.abstractmethod
    def __init__(self, dir, url):
        ...

    @property
    @abc.abstractmethod
    def directory(self):
        ...

    @property
    @abc.abstractmethod
    def url(self):
        ...

    @abc.abstractmethod
    def clone(self, branch):
        ...

    @abc.abstractmethod
    def pull(self, branch):
        ...

    @abc.abstractmethod
    def push(self, branch):
        ...

    @abc.abstractmethod
    def commit(self, message, *pathnames):
        ...



import os

import pytest
import sh

from .git import Git
from .hg import Hg


def populate_repository(d):
    open(f"{d}/README.md", "w").write("# testrepo")
    open(f"{d}/info.txt", "w").write("# someinfo")


def git_repository(d):
    sh.git.init(d)
    populate_repository(d)
    sh.git("-C", d, "add", "README.md", "info.txt")
    sh.git("-C", d, "commit", "-m", "initial")
    sh.git("-C", d, "checkout", "-b", "otherbranch")


def hg_repository(d):
    sh.hg.init(d)
    open(f"{d}/.hgkeep", "w").write("# init")
    sh.hg("--cwd", d, "add", ".hgkeep")
    sh.hg("--cwd", d, "commit", "-m", "initial")

    sh.hg("--cwd", d, "branch", "master")
    populate_repository(d)
    sh.hg("--cwd", d, "add", "README.md", "info.txt")
    sh.hg("--cwd", d, "commit", "-m", "initial")

    sh.hg("--cwd", d, "branch", "otherbranch")


@pytest.fixture(
    params=[
        (Git, git_repository),
        (Hg, hg_repository),
    ],
    ids=[
        "Git",
        "Hg",
    ],
)
def dvcs(request, tmpdir):
    (DVCS, repository) = request.param
    directory = f"{tmpdir}/testrepository"
    origin = f"{tmpdir}/origin"
    repository(origin)
    return DVCS(dir=directory, url=origin)


def test_dvcs_properties(dvcs):
    assert dvcs.directory
    assert dvcs.url


def test_dvcs_clone_master(dvcs):
    assert dvcs.clone("master") is True
    assert os.path.exists(f"{dvcs.directory}/README.md")
    assert dvcs.clone("master") is False


def test_dvcs_clone_not_master(dvcs):
    assert dvcs.clone("other") is True
    assert not os.path.exists(f"{dvcs.directory}/README.md")
    assert dvcs.clone("other") is False


def test_dvcs_commit(dvcs, tmpdir):
    assert dvcs.clone("master") is True
    content = "C"
    open(f"{dvcs.directory}/README.md", "w").write(content)
    dvcs.commit("message", "README.md")
    dvcs.push("master")

    DVCS = type(dvcs)
    other_dvcs = DVCS(dir=f"{tmpdir}/other", url=dvcs.url)
    assert other_dvcs.clone("master") is True
    assert open(f"{other_dvcs.directory}/README.md").read() == content


def test_dvcs_push_pull(dvcs, tmpdir):
    assert dvcs.clone("master") is True

    DVCS = type(dvcs)
    other_dvcs = DVCS(dir=f"{tmpdir}/other", url=dvcs.url)
    assert other_dvcs.clone("master") is True
    content = "C"
    open(f"{other_dvcs.directory}/README.md", "w").write(content)
    other_dvcs.commit("message", "README.md")
    other_dvcs.push("master")

    dvcs.pull("master")
    assert open(f"{dvcs.directory}/README.md").read() == content



import os

import pytest

from .fedeproxy import Fedeproxy


@pytest.mark.forges(["GitLab"])
def test_fedeproxy_init(forge):
    Forge = type(forge)
    f = Fedeproxy(Forge, forge.url)
    f.init()
    assert isinstance(f.own, Forge)
    assert isinstance(f.forge, Forge)


@pytest.mark.forges(["GitLab"])
def test_fedeproxy_export(tmpdir, forge, make_user, password, make_project):
    f = Fedeproxy(type(forge), forge.url)
    f.base_directory = f"{tmpdir}/fedeproxy"
    f.init()

    username = "testuser1"
    make_user(username)
    f.forge.authenticate(username=username, password=password)
    make_project(username, "testproject")

    make_user(f.own.owner)
    f.own.authenticate(username=f.own.owner, password=password)
    make_project(f.own.namespace, "fedeproxy")

    pathname = f.project_export(username, "testproject")
    os.path.exists(pathname)



import hashlib

from fedeproxy.interfaces import forge

from .identity import Identities


class Fedeproxy(object):
    def __init__(self, forge_factory, url):
        self._forge_factory = forge_factory
        self._url = url
        self.base_directory = None

    def init(self):
        self.own_forge_create()
        self.forge_create()

    @property
    def url(self):
        return self._url

    @property
    def own(self):
        return self._own

    def own_forge_create(self):
        class F(FedeproxyOwnForge, self._forge_factory):
            pass

        self._own = F(self.url)
        self.own.base_directory = self.base_directory

    @property
    def forge(self):
        return self._forge

    def forge_create(self):
        class F(FedeproxyForge, self._forge_factory):
            pass

        self._forge = F(self.url)

    def project_export(self, namespace, project):
        p = self.forge.project_create(namespace, project)
        d = self.own.project_create().dvcs()
        d.clone(p.name)
        p.save(d.directory)
        d.commit("exported project", "issues.json")
        d.push(p.name)
        return f"{d.directory}/issues.json"


class FedeproxyOwnForge(forge.Forge):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._namespace = "fedeproxy"
        self._owner = "fedeproxy"

    def init(self):
        self.identities = Identities(self)
        self.identities.init()

    @property
    def owner(self):
        return self._owner

    @property
    def namespace(self):
        return self._namespace

    def project_delete(self):
        return super().project_delete(self.namespace, "fedeproxy")

    def project_factory(self):
        class F(FedeproxyOwnProject, super().project_factory()):
            pass

        return F

    def project_create(self):
        return super().project_create(self.namespace, "fedeproxy")


class FedeproxyOwnProject(forge.Project):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def dvcs(self):
        return super().dvcs(f"{self.forge.base_directory}/fedeproxy")


class FedeproxyForge(forge.Forge):
    def project_factory(self):
        class F(FedeproxyProject, super().project_factory()):
            pass

        return F


class FedeproxyProject(forge.Project):
    @property
    def name(self):
        return hashlib.sha256(self.http_url_to_repo.encode("ascii")).hexdigest()



from allauth.socialaccount import models
from django.conf import settings
from rest_framework import permissions

from fedeproxy.common.gitlab import GitLab


#
# The only justification for making this function instead of using the method is
# for testing purposes. It is really difficult to mock the IsFedeproxyGroupMember.has_permission
# because the contet in which IsFedeproxyGroupMember is instantiated is hard to predict.
#
def has_permission(request, view):
    account = models.SocialAccount.objects.get(user=request.user)
    social_token = models.SocialToken.objects.get(account=account)
    gitlab = GitLab(settings.FEDEPROXY_GITLAB_URL)
    gitlab.set_token(social_token.token)
    return gitlab.is_self_member_of_group("fedeproxy")


class IsFedeproxyGroupMember(permissions.BasePermission):
    def has_permission(self, request, view):
        return has_permission(request, view)



"""fedeproxy URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.conf.urls import include, url
from django.conf.urls.static import static
from django.urls import path
from django.views.generic import TemplateView

from fedeproxy.api import views

from .views import member_index

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    path("ping/", views.ping, name="ping"),
    url(r"^$", TemplateView.as_view(template_name="visitor/landing-index.html"), name="landing_index"),
    url(r"^accounts/", include("allauth.urls")),
    url(r"^member/$", member_index, name="user_home"),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



import os

from .gitea import Gitea


def test_gitea_fork():
    (Forge, url) = (Gitea, f"http://{os.environ.get('MY_IP', '0.0.0.0')}:8781")
    forge = Forge(url)
    forge.authenticate(username="root", password="Wrobyak4")
    password = "Wrobyak4"

    username1 = "testuser1"
    email1 = "testuser1@example.com"
    forge.user_create(username1, password, email1)
    forge.project_delete(username1, "testproject")

    username2 = "testuser2"
    email2 = "testuser2@example.com"
    forge.user_create(username2, password, email2)
    forge.project_delete(username2, "testproject")

    forge.authenticate(username=username1, password=password)
    p1 = forge.project_create(username1, "testproject")

    forge.authenticate(username=username2, password=password)
    p2 = forge.project_fork(username1, "testproject")
    assert p1.id != p2.id
    assert p2.project == "testproject"
    assert p1.project == p2.project

    p3 = forge.project_fork(username1, "testproject")
    assert p2.id == p3.id

    forge.authenticate(username="root", password="Wrobyak4")
    assert forge.project_delete(username2, "testproject") is True
    assert forge.user_delete(username2) is True
    assert forge.project_delete(username1, "testproject") is True
    assert forge.user_delete(username1) is True



import pytest

from . import retry


def test_that_retry_works_on_simple_function():
    @retry.retry(Exception)
    def f():
        return True

    assert f()


def test_that_retry_fails():
    @retry.retry(AssertionError)
    def f():
        assert 0

    with pytest.raises(retry.RetryException):
        f()


def test_that_retry_works_on_complex_function():
    class C:
        fail = 2

        @retry.retry(Exception, tries=3)
        def f(self):
            self.fail -= 1
            assert self.fail == 0
            return True

    c = C()
    assert c.f()
    assert c.fail == 0



import pytest
from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient, APITestCase

UserModel = get_user_model()


class APIUserAPITestCase(APITestCase):
    @pytest.mark.django_db
    def setUp(self):
        self.user = UserModel.objects.create_user(username="test", email="test@...", password="top_secret")
        token = Token.objects.create(user=self.user)
        self.client = APIClient()
        self.client.credentials(HTTP_AUTHORIZATION="Token " + token.key)


class TestAPI(APIUserAPITestCase):
    @pytest.mark.django_db
    def test_ping(self):
        url = reverse("ping")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"pong")



import os

import sh

from fedeproxy.interfaces.dvcs import DVCS


class Hg(DVCS):
    def __init__(self, dir, url):
        self._url = url
        self.d = dir
        self.h = sh.hg.bake("--cwd", self.d)

    @property
    def directory(self):
        return self.d

    @property
    def url(self):
        return self._url

    def clone(self, branch):
        clone = not os.path.exists(self.d)
        if clone:
            sh.hg.clone(self.url, self.d)
        try:
            self.h.update(branch)
        except sh.ErrorReturnCode_255:
            self._create_branch(branch)
        return clone

    def pull(self, branch):
        self.h.pull("--branch", branch)
        self.h.update(branch)

    def push(self, branch):
        self.h.push("--branch", branch)

    def commit(self, message, *pathnames):
        self.h.add(*pathnames)
        self.h.commit("-m", message)

    def _branch_exists(self, branch):
        try:
            self.h.branches("--rev", branch)
            return True
        except sh.ErrorReturnCode_255:
            return False

    def _create_branch(self, branch):
        if self._branch_exists(branch):
            return False
        self.h.update("--rev", "0")
        self.h.branch(branch)
        self.h.commit("-m" f"create branch {branch}")
        self.h.push("--new-branch", "--branch", branch)
        return True



import json
import logging
import os

from jschon import JSON, Catalogue, JSONSchema

logger = logging.getLogger(__name__)

Catalogue.create_default_catalogue("2020-12")


class LoadError(Exception):
    pass


class Format(object):
    def __init__(self):
        self.d = f"{os.path.dirname(__file__)}/data"
        self.schema = None

    def load(self, filename):
        return json.load(open(filename))

    def save(self, filename, data):
        json.dump(data, open(filename, "w"))

    def validate(self, data):
        j = JSON(data)
        r = self.schema.evaluate(j)
        if r.valid:
            return True
        logger.error(r.output("verbose"))
        raise LoadError(j.path)


class FormatIssue(Format):
    def __init__(self):
        super().__init__()
        JSONSchema.loadf(f"{self.d}/comment.json")
        self.schema = JSONSchema.loadf(f"{self.d}/issue.json")



"""
Django settings for fedeproxy project.

Generated by 'django-admin startproject' using Django 2.1.5.

For more information on this file, see
https://docs.djangoproject.com/en/2.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.1/ref/settings/
"""

import os
from os.path import dirname, join

# Build paths inside the project like this: os.path.join(CONFIG_DIR, ...)
FEDEPROXY_DOT = os.environ.get("FEDEPROXY_DOT", os.path.expanduser("~/.fedeproxy"))

if os.path.exists("setup.cfg"):  # running from sources
    CONFIG_DIR = os.path.abspath(".")
else:
    CONFIG_DIR = f"{FEDEPROXY_DOT}"

CERTS_DIR = f"{CONFIG_DIR}/certs"

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "mhh^de)&ky$&g083tldl@d7km-6)jc=cygb+07"

# SECURITY WARNING: don't run with debug turned on in production!
# DEBUG = True

ALLOWED_HOSTS = ["*"]


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "rest_framework",
    "rest_framework.authtoken",
    "fedeproxy.api",
    "bootstrap3",
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    "allauth.socialaccount.providers.gitlab",  # enabled by configure
    "fedeproxy.auth",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "fedeproxy.urls"

AUTHENTICATION_BACKENDS = ("allauth.account.auth_backends.AuthenticationBackend",)

MODULE_DIR = dirname(__file__)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            join(MODULE_DIR, "templates", "allauth"),
            join(MODULE_DIR, "templates"),
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.template.context_processors.static",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "fedeproxy.wsgi.application"


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(CONFIG_DIR, "db.sqlite3"),
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = "/static/"

STATICFILES_DIRS = (join(MODULE_DIR, "static"),)

REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 10,
}

FEDEPROXY_GITLAB_URL = "https://???"

SITE_ID = 1
ACCOUNT_DEFAULT_HTTP_PROTOCOL = os.environ.get("ACCOUNT_DEFAULT_HTTP_PROTOCOL", "https")
LOGIN_REDIRECT_URL = "/member/"
ACCOUNT_AUTHENTICATION_METHOD = "username"
ACCOUNT_LOGOUT_ON_GET = True
ACCOUNT_USERNAME_REQUIRED = False
ACCOUNT_EMAIL_REQUIRED = False
ACCOUNT_USERNAME_MIN_LENGTH = 1
ACCOUNT_EMAIL_VERIFICATION = "none"  # testing...
# ACCOUNT_USER_MODEL_USERNAME_FIELD = None
SOCIALACCOUNT_AUTO_SIGNUP = False  # require social accounts to use the signup form ... I think
# For custom sign-up form:
# http://stackoverflow.com/questions/12303478/how-to-customize-user-profile-when-using-django-allauth
SOCIALACCOUNT_PROVIDERS = {
    "gitlab": {
        "GITLAB_URL": FEDEPROXY_GITLAB_URL,
        "SCOPE": ["read_user", "api"],
    },
}



import logging
import time
from functools import wraps

logger = logging.getLogger(__name__)


class RetryException(Exception):
    pass


def retry(exceptions, tries=2, delay=1):
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries + 1, delay
            while mtries > 0:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    logger.info("%s: %s, Retrying in %s seconds...", f.__qualname__, e, mdelay)
                    if mtries == tries + 1:
                        logger.debug("", exc_info=True)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= 2
            raise RetryException("Number of retries exceeded for function " + f.__name__)

        return f_retry

    return deco_retry



import os
from pathlib import Path

import git
import gitdb
import sh

from fedeproxy.interfaces.dvcs import DVCS


class Git(DVCS):
    def __init__(self, dir, url):
        self._url = url
        self.d = dir
        self.g = sh.git.bake("-C", self.d)

    @property
    def directory(self):
        return self.d

    @property
    def url(self):
        return self._url

    def clone(self, branch):
        clone = not os.path.exists(self.d)
        if clone:
            sh.git.clone(self.url, self.d)
        try:
            self.g.checkout(branch)
        except sh.ErrorReturnCode_1:
            self._create_branch(branch)
        return clone

    def _branch_exists(self, branch):
        r = git.Repo(self.d)
        try:
            r.commit(f"origin/{branch}")
            return True
        except gitdb.exc.BadName:
            return False

    def _create_branch(self, branch):
        if self._branch_exists(branch):
            return False
        self.g.checkout("--orphan", branch)
        gitkeep = f"{self.d}/.gitkeep"
        Path(gitkeep).touch()
        self.g.add(gitkeep)
        self.g.commit("-m", ".gitkeep", gitkeep)
        self.g.push("origin", branch)
        self.g.reset("--hard", branch)
        return True

    def pull(self, branch):
        self.g.checkout(branch)
        self.g.pull()

    def push(self, branch):
        self.g.checkout(branch)
        self.g.push("origin", branch)

    def commit(self, message, *pathnames):
        self.g.add(*pathnames)
        self.g.commit("-m", message)



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.template import loader
from rest_framework.authtoken.models import Token


@login_required
def member_index(request):
    t = loader.get_template("member/member-index.html")
    token = Token.objects.get_or_create(user=request.user)
    c = {"token": token[0]}
    return HttpResponse(t.render(c, request), content_type="text/html")



import os

import pytest

from fedeproxy.architecture.forge.gitea import Gitea
from fedeproxy.architecture.forge.gitlab import GitLab

gitea_url = f"http://{os.environ.get('MY_IP', '0.0.0.0')}:8781"
gitlab_url = f"http://{os.environ.get('MY_IP', '0.0.0.0')}:8181"


@pytest.fixture(
    params=[
        (Gitea, gitea_url),
        (GitLab, gitlab_url),
    ],
    ids=[
        Gitea.__name__,
        GitLab.__name__,
    ],
)
def forge(request):
    marker = request.node.get_closest_marker("forges")
    (Forge, url) = request.param
    if marker is not None:
        forges = marker.args[0]
        if Forge.__name__ not in forges:
            pytest.skip(f"the test has a 'mark' excluding {Forge.__name__} ")
    return Forge(url)


@pytest.fixture
def password():
    return "Wrobyak4"


@pytest.fixture
def make_user(forge, password):

    usernames = []

    def _make_user(username):
        forge.authenticate(username="root", password="Wrobyak4")
        forge.user_delete(username)
        email = f"{username}@example.com"
        forge.user_create(username, password, email)
        usernames.append(username)

    yield _make_user

    forge.authenticate(username="root", password="Wrobyak4")
    for username in usernames:
        forge.user_delete(username)


@pytest.fixture
def make_project(forge, password):
    projects = []

    def _make_project(username, project):
        forge.project_delete(username, project)
        p = forge.project_create(username, project)
        projects.append((username, project))
        return p

    yield _make_project

    forge.authenticate(username="root", password=password)
    for project in projects:
        forge.project_delete(*project)



import pytest

from . import format


@pytest.fixture
def issue():
    return {
        "url": "http://example.com/issue",
        "title": "title",
        "description": "description",
        "repository": "http://example.com/repository",
        "user": "http://example.com/issue-user",
        "created_at": "2016-01-04T15:31:51.081Z",
        "closed_at": "2021-01-02 10:30:50",
        "state": "closed",
        "comments": [
            {
                "url": "http://example.com/comment1",
                "body": "body1",
                "user": "http://example.com/comment-user1",
                "created_at": "2021-01-01 10:30:50",
                "id": "comment1",
            },
            {
                "url": "http://example.com/comment2",
                "body": "body2",
                "user": "http://example.com/comment-user2",
                "created_at": "2021-01-01 10:30:50",
                "id": "comment2",
            },
        ],
        "id": "issue1234",
    }


def test_format_issue(issue, tmpdir):
    f = format.FormatIssue()
    assert f.schema.validate().valid
    p = f"{tmpdir}/issue.json"
    f.save(p, issue)
    assert f.load(p) == issue
    assert f.validate(issue)



import logging
import os
import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

from fedeproxy.version import __version__

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "fedeproxy.settings")


class FedeproxyApp(App):
    def __init__(self, modules):
        super().__init__(
            description="fedeproxy",
            version=__version__,
            command_manager=CommandManager(modules),
            deferred_help=True,
        )

    def configure_logging(self):
        super().configure_logging()

        root_logger = logging.getLogger("")
        root_logger.setLevel(logging.WARNING)

        level = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}.get(
            self.options.verbose_level, logging.DEBUG
        )

        root_logger = logging.getLogger("fedeproxy")
        root_logger.setLevel(level)


def main(argv=sys.argv[1:]):
    myapp = FedeproxyApp("fedeproxy.cli")
    return myapp.run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import copy
import logging
import time

import requests

from fedeproxy.common.retry import retry

from . import forge

logger = logging.getLogger(__name__)


class Gitea(forge.Forge):
    def __init__(self, url):
        super().__init__(url)
        self._s = requests.Session()

    def authenticate(self, **kwargs):
        self._session()
        if "token" in kwargs:
            self.set_token(kwargs["token"])
        else:
            self.login(kwargs["username"], kwargs["password"])
        self._user = self.s.get(f"{self.s.api}/user").json()

    @property
    def username(self):
        return self._user["login"]

    @property
    def is_admin(self):
        return self._user["is_admin"]

    def _session(self):
        self.s.api = f"{self.url}/api/v1"

    def login(self, username, password):
        r = self.s.post(
            f"{self.s.api}/users/{username}/tokens",
            auth=(username, password),
            json={
                "name": f"TEST{time.time()}",
            },
        )
        r.raise_for_status()
        self.set_token(r.json()["sha1"])

    def _finalize_user_create(self, username, password):
        r = self.s.post(
            f"{self.url}/user/login",
            data={
                "user_name": username,
                "password": password,
            },
        )
        r.raise_for_status()
        r = self.s.post(
            f"{self.url}/user/settings/change_password",
            data={
                "password": password,
                "retype": password,
                "_csrf": self.s.cookies["_csrf"],
            },
        )
        r.raise_for_status()

    def get_token(self):
        return self.token

    def set_token(self, token):
        self.token = token
        self.s.headers["Authorization"] = f"token {token}"

    def user_delete(self, username):
        user = self.user_get(username)
        if user is None:
            return False
        while True:
            r = self.s.delete(f"{self.s.api}/admin/users/{username}")
            if r.status_code == 404:
                break
            logger.debug(r.text)
            r.raise_for_status()
        return True

    def user_factory(self):
        return GiteaUser

    def user_get(self, username):
        if username == self.username:
            r = self.s.get(f"{self.s.api}/user")
        elif self.is_admin:
            r = self.s.get(f"{self.s.api}/user", params={"sudo": username})
        else:
            r = self.s.get(f"{self.s.api}/users/{username}")
        if r.status_code == 404:
            return None
        else:
            r.raise_for_status()
            return self.user_factory()(self, r.json())

    def user_create(self, username, password, email):
        info = self.user_get(username)
        if info is None:
            r = self.s.post(
                f"{self.s.api}/admin/users",
                data={
                    "username": username,
                    "email": email,
                    "password": password,
                },
            )
            logger.debug(r.text)
            r.raise_for_status()
            info = r.json()
            Gitea(self.url)._finalize_user_create(username, password)
            info = self.user_get(username)
        return info

    def project_delete(self, namespace, project):
        p = self.project_get(namespace, project)
        if p is None:
            return False
        r = self.s.delete(f"{self.s.api}/repos/{namespace}/{project}")
        r.raise_for_status()
        while self.project_get(namespace, project) is not None:
            time.sleep(1)
        return True

    def project_factory(self):
        return GiteaProject

    def project_get(self, namespace, project):
        r = self.s.get(f"{self.s.api}/repos/{namespace}/{project}")
        if r.status_code == requests.codes.ok:
            return self.project_factory()(self, r.json())
        else:
            return None

    class ForkingInProgress(Exception):
        pass

    @retry(ForkingInProgress, tries=5)
    def _project_wait_fork(self, namespace, project):
        p = self.project_get(namespace, project)
        if p is None:
            raise Gitea.ForkingInProgress(f"Waiting for {namespace}/{project} fork to finish.")
        return p

    def project_fork(self, namespace, project, target_namespace=None):
        data = {}
        if target_namespace is None:
            target_namespace = self.username
        else:
            data["organization"] = target_namespace
        p = self.project_get(target_namespace, project)
        if p is not None:
            return p
        r = self.s.post(f"{self.s.api}/repos/{namespace}/{project}/forks", json=data)
        r.raise_for_status()
        return self._project_wait_fork(target_namespace, project)

    class DeletionInProgress(Exception):
        pass

    @retry(DeletionInProgress, tries=5)
    def _project_create(self, namespace, project, **data):
        data.update(
            {
                "name": project,
            }
        )
        r = self.s.post(f"{self.s.api}/user/repos", data=data)
        logger.info(r.text)
        if r.status_code == 201:
            return self.project_get(namespace, project)
        r.raise_for_status()

    def project_create(self, namespace, project, **data):
        p = self.project_get(namespace, project)
        if p is None:
            return self._project_create(namespace, project, **data)
        else:
            return p


class GiteaProject(forge.Project):
    @property
    def id(self):
        return self._project["id"]

    @property
    def namespace(self):
        return self._project["owner"]["login"]

    @property
    def project(self):
        return self._project["name"]

    def milestones_factory(self):
        return GiteaMilestones

    def issues_factory(self):
        return GiteaIssues

    @property
    def ssh_url_to_repo(self):
        return self._project["ssh_url"]

    @property
    def http_url_to_repo(self):
        return self._project["clone_url"]


class GiteaMilestones(forge.Milestones):
    def _build_url(self):
        return f"{self.s.api}/repos/{self.project.namespace}/{self.project.project}/milestones"

    def get(self, id):
        r = self.s.get(f"{self._build_url()}/{id}")
        if r.status_code == requests.codes.ok:
            return GiteaMilestone(self.project, r.json())
        else:
            return None

    def delete(self, id):
        p = self.get(id)
        if p is None:
            return False
        r = self.s.delete(f"{self._build_url()}/{id}")
        r.raise_for_status()
        return True

    def create(self, title, **data):
        data = copy.copy(data)
        data.update({"title": title})
        r = self.s.post(self._build_url(), data=data)
        logger.info(r.text)
        if r.status_code == 201:
            m = r.json()
            return self.get(m["id"])
        r.raise_for_status()

    def list(self):
        pass


class GiteaMilestone(forge.Milestone):
    @property
    def id(self):
        return self._milestone["id"]


class GiteaIssues(forge.Issues):
    def get(self, id):
        pass

    def delete(self, id):
        pass

    def create(self, title, **data):
        pass

    def list(self):
        pass


class GiteaIssue(forge.Issue):
    @property
    def id(self):
        return self._issue["iid"]

    def to_json(self):
        pass

    def from_json(self, j):
        pass


class GiteaUser(forge.User):
    @property
    def url(self):
        return f"{self.forge.url}/{self.username}"

    @property
    def username(self):
        return self._user["username"]

    @property
    def emails(self):
        return [self._user["email"]]



import ndjson
from furl import furl

from ...format import format
from ...interfaces import forge
from ..dvcs import git


class Forge(forge.Forge):
    pass


class Project(forge.Project):
    def save(self, pathname):
        self.issues.save(f"{pathname}/issues.json")

    def load(self, pathname):
        return self.issues.load(f"{pathname}/issues.json")

    def dvcs_factory(self):
        return git.Git

    def dvcs(self, directory):
        o = furl(self.http_url_to_repo)
        o.username = "oauth2"
        o.password = self.forge.get_token()
        return self.dvcs_factory()(directory, o.tostr())


class Milestones(forge.Milestones):
    pass


class Milestone(forge.Milestone):
    pass


class Issues(forge.Issues):
    def save(self, pathname):
        with open(pathname, "w") as f:
            writer = ndjson.writer(f, ensure_ascii=False)
            for i in self.list():
                writer.writerow(i.to_json())

    def load(self, pathname):
        fi = format.FormatIssue()
        count = 0
        with open(pathname) as f:
            for j in ndjson.reader(f):
                fi.validate(j)
                self.create_from_json(j)
                count += 1
        return count

    def create_from_json(self, j):
        i = self.get(j["id"])
        if i is None:
            i = self.create(j["title"])
        i.from_json(j)


class Issue(forge.Issue):
    pass


class User(forge.User):
    pass



import hashlib
import json
import os


class Identity(object):
    def __init__(self):
        self.emails = []
        self.url = None
        self.owned = False

    def is_staged(self):
        return len(self.emails) == 0

    def is_owned(self):
        return self.owned

    def serialize(self):
        return self.__dict__

    def unserialize(self, d):
        self.__dict__.update(d)
        return self


class IdentityPrivate(Identity):
    def __init__(self):
        super().__init__()
        self.token = None


class IdentityPublic(Identity):
    def from_private(self, identity):
        self.emails = [hashlib.sha256(e.encode("UTF-8")).hexdigest() for e in identity.emails]
        self.url = identity.url
        self.owned = identity.owned


class Identities(object):
    def __init__(self, forge):
        self.forge = forge
        self.email = None
        self.identities = []

    def init(self):
        self.d = self.forge.project_create().dvcs()
        self.p = f"{self.d.directory}/identities.json"
        self.load()

    def load(self):
        self.d.clone("identities")
        if os.path.exists(self.p):
            for i in json.load(open(self.p)):
                self.identities.append(self.IdentityClass().unserialize(i))
            return True
        else:
            return False

    def save(self):
        self.identities = [i.serialize() for i in self.identities]
        json.dump(self.identities, open(self.p, "w"))


class IdentitiesPrivate(Identities):

    IdentityClass = IdentityPrivate


class IdentitiesPublic(Identity):

    IdentityClass = IdentityPublic
